package k4unl.minecraft.sip.network.rcon;

import k4unl.minecraft.sip.lib.Log;
import k4unl.minecraft.sip.network.TCPServerThread;
import net.minecraft.network.rcon.IServer;
import net.minecraft.network.rcon.RConOutputStream;
import net.minecraft.network.rcon.RConUtils;
import net.minecraft.server.MinecraftServer;

import java.io.IOException;
import java.net.*;
import java.util.*;

/**
 * @author Koen Beckers (K-4U)
 */
public class RConThreadQuery extends net.minecraft.network.rcon.RConThreadQuery {
    
    /**
     * The time of the last client auth check
     */
    private long           lastAuthCheckTime;
    /**
     * The RCon query port
     */
    private int            queryPort;
    /**
     * Port the server is running on
     */
    private int            serverPort;
    /**
     * The maximum number of players allowed on the server
     */
    private int            maxPlayers;
    /**
     * The current server message of the day
     */
    private String         serverMotd;
    /**
     * The name of the currently loaded world
     */
    private String         worldName;
    /**
     * The remote socket querying the server
     */
    private DatagramSocket querySocket;
    /**
     * A buffer for incoming DatagramPackets
     */
    private byte[] buffer = new byte[1460];
    /**
     * Storage for incoming DatagramPackets
     */
    private DatagramPacket   incomingPacket;
    private Map              field_72644_p;
    /**
     * The hostname of this query server
     */
    private String           queryHostname;
    /**
     * The hostname of the running server
     */
    private String           serverHostname;
    /**
     * A map of SocketAddress objects to RConThreadQueryAuth objects
     */
    private Map              queryClients;
    /**
     * The time that this RConThreadQuery was constructed, from (new Date()).getTime()
     */
    private long             time;
    /**
     * The RConQuery output stream
     */
    private RConOutputStream output;
    /**
     * The time of the last query response sent
     */
    private long             lastQueryResponseTime;
    private static final String __OBFID = "CL_00001802";
    
    public RConThreadQuery(IServer p_i1536_1_) {
        
        super(p_i1536_1_);
        this.queryPort = p_i1536_1_.func_71327_a("query.port", 0);
        this.serverHostname = p_i1536_1_.func_71277_t();
        this.serverPort = p_i1536_1_.func_71234_u();
        this.serverMotd = p_i1536_1_.func_71274_v();
        this.maxPlayers = p_i1536_1_.func_71275_y();
        this.worldName = p_i1536_1_.func_71270_I();
        this.lastQueryResponseTime = 0L;
        this.queryHostname = "0.0.0.0";
        
        if (0 != this.serverHostname.length() && !this.queryHostname.equals(this.serverHostname)) {
            this.queryHostname = this.serverHostname;
        } else {
            this.serverHostname = "0.0.0.0";
            
            try {
                InetAddress inetaddress = InetAddress.getLocalHost();
                this.queryHostname = inetaddress.getHostAddress();
            } catch (UnknownHostException unknownhostexception) {
                this.func_72606_c("Unable to determine local host IP, please set server-ip in \'" + p_i1536_1_.func_71329_c() + "\' : " + unknownhostexception.getMessage());
            }
        }
        
        if (0 == this.queryPort) {
            this.queryPort = this.serverPort;
            this.func_72609_b("Setting default query port to " + this.queryPort);
            p_i1536_1_.func_71328_a("query.port", Integer.valueOf(this.queryPort));
            p_i1536_1_.func_71328_a("debug", Boolean.valueOf(false));
            p_i1536_1_.func_71326_a();
        }
        
        this.field_72644_p = new HashMap();
        this.output = new RConOutputStream(1460);
        this.queryClients = new HashMap();
        this.time = (new Date()).getTime();
    }
    
    /**
     * Sends a byte array as a DatagramPacket response to the client who sent the given DatagramPacket
     */
    private void sendResponsePacket(byte[] p_72620_1_, DatagramPacket p_72620_2_) throws IOException {
        
        this.querySocket.send(new DatagramPacket(p_72620_1_, p_72620_1_.length, p_72620_2_.getSocketAddress()));
    }
    
    /**
     * Parses an incoming DatagramPacket, returning true if the packet was valid
     */
    private boolean parseIncomingPacket(DatagramPacket p_72621_1_) throws IOException {
        
        byte[] abyte = p_72621_1_.getData();
        int i = p_72621_1_.getLength();
        SocketAddress socketaddress = p_72621_1_.getSocketAddress();
        this.func_72607_a("Packet len " + i + " [" + socketaddress + "]");
        
        if (3 <= i && -2 == abyte[0] && -3 == abyte[1]) {
            this.func_72607_a("Packet \'" + RConUtils.func_72663_a(abyte[2]) + "\' [" + socketaddress + "]");
            
            switch (abyte[2]) {
                case 8:
                    if (!this.verifyClientAuth(p_72621_1_)) {
                        this.func_72607_a("Invalid challenge [" + socketaddress + "]");
                        return false;
                    } else {
                        this.func_72607_a("Asked for the special packet!");
                        
                        RConOutputStream outputStream = new RConOutputStream(1460);
                        outputStream.func_72667_a(0);
                        outputStream.func_72670_a(this.getRequestID(p_72621_1_.getSocketAddress()));
                        //Tell them the port we're listening on.
                        outputStream.func_72671_a(TCPServerThread.getPort() + "");
    
                        this.sendResponsePacket(outputStream.func_72672_a(), p_72621_1_);
                        this.func_72607_a("Status [" + socketaddress + "]");
                        
                        this.sendAuthChallenge(p_72621_1_);
                        this.func_72607_a("Challenge [" + socketaddress + "]");
                        return true;
                    }
                case 0:
                    if (!this.verifyClientAuth(p_72621_1_)) {
                        this.func_72607_a("Invalid challenge [" + socketaddress + "]");
                        return false;
                    } else if (15 == i) {
                        this.sendResponsePacket(this.createQueryResponse(p_72621_1_), p_72621_1_);
                        this.func_72607_a("Rules [" + socketaddress + "]");
                    } else {
                        RConOutputStream rconoutputstream = new RConOutputStream(1460);
                        rconoutputstream.func_72667_a(0);
                        rconoutputstream.func_72670_a(this.getRequestID(p_72621_1_.getSocketAddress()));
                        rconoutputstream.func_72671_a(this.serverMotd);
                        rconoutputstream.func_72671_a("SMP");
                        rconoutputstream.func_72671_a(this.worldName);
                        rconoutputstream.func_72671_a(Integer.toString(this.func_72603_d()));
                        rconoutputstream.func_72671_a(Integer.toString(this.maxPlayers));
                        rconoutputstream.func_72668_a((short) this.serverPort);
                        rconoutputstream.func_72671_a(this.queryHostname);
                        this.sendResponsePacket(rconoutputstream.func_72672_a(), p_72621_1_);
                        this.func_72607_a("Status [" + socketaddress + "]");
                    }
                case 9:
                    this.sendAuthChallenge(p_72621_1_);
                    this.func_72607_a("Challenge [" + socketaddress + "]");
                    return true;
                default:
                    return true;
            }
        } else {
            this.func_72607_a("Invalid packet [" + socketaddress + "]");
            return false;
        }
    }
    
    
    /**
     * Creates a query response as a byte array for the specified query DatagramPacket
     */
    private byte[] createQueryResponse(DatagramPacket p_72624_1_) throws IOException {
        
        long i = MinecraftServer.func_130071_aq();
        
        if (i < this.lastQueryResponseTime + 5000L) {
            byte[] abyte = this.output.func_72672_a();
            byte[] abyte1 = this.getRequestID(p_72624_1_.getSocketAddress());
            abyte[1] = abyte1[0];
            abyte[2] = abyte1[1];
            abyte[3] = abyte1[2];
            abyte[4] = abyte1[3];
            return abyte;
        } else {
            this.lastQueryResponseTime = i;
            this.output.func_72669_b();
            this.output.func_72667_a(0);
            this.output.func_72670_a(this.getRequestID(p_72624_1_.getSocketAddress()));
            this.output.func_72671_a("splitnum");
            this.output.func_72667_a(128);
            this.output.func_72667_a(0);
            this.output.func_72671_a("hostname");
            this.output.func_72671_a(this.serverMotd);
            this.output.func_72671_a("gametype");
            this.output.func_72671_a("SMP");
            this.output.func_72671_a("game_id");
            this.output.func_72671_a("MINECRAFT");
            this.output.func_72671_a("version");
            this.output.func_72671_a(this.field_72617_b.func_71249_w());
            this.output.func_72671_a("plugins");
            this.output.func_72671_a(this.field_72617_b.func_71258_A());
            this.output.func_72671_a("map");
            this.output.func_72671_a(this.worldName);
            this.output.func_72671_a("numplayers");
            this.output.func_72671_a("" + this.func_72603_d());
            this.output.func_72671_a("maxplayers");
            this.output.func_72671_a("" + this.maxPlayers);
            this.output.func_72671_a("hostport");
            this.output.func_72671_a("" + this.serverPort);
            this.output.func_72671_a("hostip");
            this.output.func_72671_a(this.queryHostname);
            this.output.func_72667_a(0);
            this.output.func_72667_a(1);
            this.output.func_72671_a("player_");
            this.output.func_72667_a(0);
            String[] astring = this.field_72617_b.func_71213_z();
            String[] astring1 = astring;
            int j = astring.length;
            
            for (int k = 0; k < j; ++k) {
                String s = astring1[k];
                this.output.func_72671_a(s);
            }
            
            this.output.func_72667_a(0);
            return this.output.func_72672_a();
        }
    }
    
    /**
     * Returns the request ID provided by the authorized client
     */
    private byte[] getRequestID(SocketAddress p_72625_1_) {
        
        return ((RConThreadQuery.Auth) this.queryClients.get(p_72625_1_)).getRequestId();
    }
    
    /**
     * Returns true if the client has a valid auth, otherwise false
     */
    private Boolean verifyClientAuth(DatagramPacket p_72627_1_) {
        
        SocketAddress socketaddress = p_72627_1_.getSocketAddress();
        
        if (!this.queryClients.containsKey(socketaddress)) {
            return Boolean.valueOf(false);
        } else {
            byte[] abyte = p_72627_1_.getData();
            return ((RConThreadQuery.Auth) this.queryClients.get(socketaddress)).getRandomChallenge() != RConUtils.func_72664_c(abyte, 7, p_72627_1_.getLength()) ? Boolean.valueOf(false) : Boolean.valueOf(true);
        }
    }
    
    /**
     * Sends an auth challenge DatagramPacket to the client and adds the client to the queryClients map
     */
    private void sendAuthChallenge(DatagramPacket p_72622_1_) throws IOException {
        
        RConThreadQuery.Auth auth = new RConThreadQuery.Auth(p_72622_1_);
        this.queryClients.put(p_72622_1_.getSocketAddress(), auth);
        this.sendResponsePacket(auth.getChallengeValue(), p_72622_1_);
    }
    
    /**
     * Removes all clients whose auth is no longer valid
     */
    private void cleanQueryClientsMap() {
        
        if (this.field_72619_a) {
            long i = MinecraftServer.func_130071_aq();
            
            if (i >= this.lastAuthCheckTime + 30000L) {
                this.lastAuthCheckTime = i;
                Iterator iterator = this.queryClients.entrySet().iterator();
                
                while (iterator.hasNext()) {
                    Map.Entry entry = (Map.Entry) iterator.next();
                    
                    if (((RConThreadQuery.Auth) entry.getValue()).hasExpired(i).booleanValue()) {
                        iterator.remove();
                    }
                }
            }
        }
    }
    
    public void run() {
        
        this.func_72609_b("Query running on " + this.serverHostname + ":" + this.queryPort);
        this.lastAuthCheckTime = MinecraftServer.func_130071_aq();
        this.incomingPacket = new DatagramPacket(this.buffer, this.buffer.length);
        
        this.field_72619_a = true;
        try {
            while (this.field_72619_a) {
                try {
                    this.querySocket.receive(this.incomingPacket);
                    this.cleanQueryClientsMap();
                    this.parseIncomingPacket(this.incomingPacket);
                } catch (SocketTimeoutException sockettimeoutexception) {
                    this.cleanQueryClientsMap();
                } catch (PortUnreachableException portunreachableexception) {
                    ;
                } catch (IOException ioexception) {
                    this.stopWithException(ioexception);
                }
            }
        } finally {
            this.func_72611_e();
        }
    }
    
    /**
     * Creates a new Thread object from this class and starts running
     */
    public void func_72602_a() {
        
        try {
            Thread.sleep(4000);
        } catch (InterruptedException e) {
            e.printStackTrace();
        }
        if (!this.field_72619_a) {
            if (0 < this.queryPort && 65535 >= this.queryPort) {
                if (this.initQuerySystem()) {
                    this.field_72618_c = new Thread(this, this.field_164003_c + " #42");
                    this.field_72618_c.start();
                    this.field_72619_a = true;
                }
            } else {
                this.func_72606_c("Invalid query port " + this.queryPort + " found in \'" + this.field_72617_b.func_71329_c() + "\' (queries disabled)");
            }
        }
    }
    
    /**
     * Stops the query server and reports the given Exception
     */
    private void stopWithException(Exception p_72623_1_) {
        
        if (this.field_72619_a) {
            this.func_72606_c("Unexpected exception, buggy JRE? (" + p_72623_1_.toString() + ")");
            
            if (!this.initQuerySystem()) {
                this.func_72610_d("Failed to recover from buggy JRE, shutting down!");
                this.field_72619_a = false;
            }
        }
    }
    
    /**
     * Initializes the query system by binding it to a port
     */
    private boolean initQuerySystem() {
        
        try {
            this.querySocket = new DatagramSocket(this.queryPort, InetAddress.getByName(this.serverHostname));
            this.func_72601_a(this.querySocket);
            this.querySocket.setSoTimeout(500);
            return true;
        } catch (SocketException socketexception) {
            this.func_72606_c("Unable to initialise query system on " + this.serverHostname + ":" + this.queryPort + " (Socket): " + socketexception.getMessage());
        } catch (UnknownHostException unknownhostexception) {
            this.func_72606_c("Unable to initialise query system on " + this.serverHostname + ":" + this.queryPort + " (Unknown Host): " + unknownhostexception.getMessage());
        } catch (Exception exception) {
            this.func_72606_c("Unable to initialise query system on " + this.serverHostname + ":" + this.queryPort + " (E): " + exception.getMessage());
        }
        
        return false;
    }
    
    @Override
    protected void func_72609_b(String p_72609_1_) {
        
        Log.info(p_72609_1_);
    }
    
    @Override
    protected void func_72607_a(String msg) {
        
        Log.debug(msg);
    }
    
    @Override
    protected void func_72606_c(String p_72606_1_) {
        
        Log.warning(p_72606_1_);
    }
    
    @Override
    protected void func_72610_d(String p_72610_1_) {
        
        Log.error(p_72610_1_);
    }
    
    class Auth {
        
        /**
         * The creation timestamp for this auth
         */
        private long timestamp = (new Date()).getTime();
        /**
         * A random integer value to be used for client response authentication
         */
        private int    randomChallenge;
        /**
         * A client-provided request ID associated with this query.
         */
        private byte[] requestId;
        /**
         * A unique string of bytes used to verify client auth
         */
        private byte[] challengeValue;
        /**
         * The request ID stored as a String
         */
        private String requestIdAsString;
        private static final String __OBFID = "CL_00001803";
        
        public Auth(DatagramPacket p_i1535_2_) {
            
            byte[] abyte = p_i1535_2_.getData();
            this.requestId = new byte[4];
            this.requestId[0] = abyte[3];
            this.requestId[1] = abyte[4];
            this.requestId[2] = abyte[5];
            this.requestId[3] = abyte[6];
            this.requestIdAsString = new String(this.requestId);
            this.randomChallenge = (new Random()).nextInt(16777216);
            this.challengeValue = String.format("\t%s%d\u0000", new Object[]{this.requestIdAsString, Integer.valueOf(this.randomChallenge)}).getBytes();
        }
        
        /**
         * Returns true if the auth's creation timestamp is less than the given time, otherwise false
         */
        public Boolean hasExpired(long p_72593_1_) {
            
            return Boolean.valueOf(this.timestamp < p_72593_1_);
        }
        
        /**
         * Returns the random challenge number assigned to this auth
         */
        public int getRandomChallenge() {
            
            return this.randomChallenge;
        }
        
        /**
         * Returns the auth challenge value
         */
        public byte[] getChallengeValue() {
            
            return this.challengeValue;
        }
        
        /**
         * Returns the request ID provided by the client.
         */
        public byte[] getRequestId() {
            
            return this.requestId;
        }
    }
}
